import markdown

DEL_RE = r'(~~)(.*?)~~'
INS_RE = r'(__)(.*?)__'
STRONG_RE = r'(\*\*)(.*?)\*\*'
EMPH_RE = r'(//)(.*?)//'

# http://achinghead.com/python-markdown-changing-bold-italics.html


class StrkUndrExtension(markdown.Extension):
    def extendMarkdown(self, md, md_globals):
        del_tag = markdown.inlinepatterns.SimpleTagPattern(DEL_RE, 'del')
        md.inlinePatterns.add('del', del_tag, '>not_strong')
        ins_tag = markdown.inlinepatterns.SimpleTagPattern(INS_RE, 'ins')
        md.inlinePatterns.add('ins', ins_tag, '>del')
        strong_tag = markdown.inlinepatterns.SimpleTagPattern(
            STRONG_RE, 'strong')
        md.inlinePatterns['strong'] = strong_tag
        emph_tag = markdown.inlinepatterns.SimpleTagPattern(EMPH_RE, 'em')
        md.inlinePatterns['emphasis'] = emph_tag
        del md.inlinePatterns['strong_em']
        del md.inlinePatterns['emphasis2']


def makeExtension(configs=None):
    return StrkUndrExtension(configs=configs)



import os
from multiprocessing import Process

import markdown
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtWebKit import QWebView, QWebPage
from whoosh import sorting
from whoosh.index import create_in, open_dir
from whoosh.qparser import QueryParser, RegexPlugin

import mikidown.mikidown_rc
from mikidown.config import __appname__, __version__
from mikidown.mikibook import NotebookListDialog
from mikidown.mikitree import *
from mikidown.mikiedit import *
from mikidown.mikiview import *
from mikidown.mikisearch import MikiSearch
from mikidown.highlighter import MikiHighlighter
from mikidown.utils import *


class MikiWindow(QMainWindow):
    def __init__(self, settings, parent=None):
        super(MikiWindow, self).__init__(parent)
        self.settings = settings
        self.notebookPath = settings.notebookPath

        self.resize(800, 600)
        screen = QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move((
            screen.width()-size.width())/2, (screen.height()-size.height())/2)
        self.setWindowTitle(
            '{} - {}'.format(settings.notebookName, __appname__))


        self.notesTree = MikiTree(self)
        self.initTree(self.notebookPath, self.notesTree)
        self.notesTree.sortItems(0, Qt.AscendingOrder)

        # Initialize whoosh index, make sure notebookPath/.indexdir exists
        self.ix = None
        indexdir = os.path.join(self.notebookPath, settings.indexdir)
        if not QDir(indexdir).exists():
            QDir().mkdir(indexdir)
            self.ix = create_in(indexdir, settings.schema)
            # Fork a process to update index, which benefit responsiveness.
            p = Process(target=self.whoosh_index, args=())
            p.start()
        else:
            self.ix = open_dir(indexdir)
        
        self.tabWidget = QTabWidget()
        self.viewedList = QToolBar(self.tr('Recently Viewed'), self)
        self.viewedList.setFixedHeight(25)
        self.notesEdit = MikiEdit(self)
        MikiHighlighter(self.notesEdit)
        self.notesView = MikiView(self)
        self.findBar = QToolBar(self.tr('Find'), self)
        self.findBar.setFixedHeight(30)
        self.noteSplitter = QSplitter(Qt.Horizontal)
        self.noteSplitter.addWidget(self.notesEdit)
        self.noteSplitter.addWidget(self.notesView)
        self.rightSplitter = QSplitter(Qt.Vertical)
        self.rightSplitter.setChildrenCollapsible(False)
        self.rightSplitter.addWidget(self.viewedList)
        self.rightSplitter.addWidget(self.noteSplitter)
        self.rightSplitter.addWidget(self.findBar)
        self.mainSplitter = QSplitter(Qt.Horizontal)
        self.mainSplitter.addWidget(self.tabWidget)
        self.mainSplitter.addWidget(self.rightSplitter)
        #self.setCentralWidget(self.mainSplitter)
        self.setCentralWidget(self.rightSplitter)
        self.mainSplitter.setStretchFactor(0, 1)
        self.mainSplitter.setStretchFactor(1, 5)

        self.searchEdit = QLineEdit()
        self.searchEdit.returnPressed.connect(self.searchNote)
        self.searchView = MikiSearch(self)
        self.searchTab = QWidget()
        searchLayout = QVBoxLayout()
        searchLayout.addWidget(self.searchEdit)
        searchLayout.addWidget(self.searchView)
        self.searchTab.setLayout(searchLayout)
        self.tocTree = QTreeWidget()
        self.tocTree.header().close()
        self.dockIndex = QDockWidget("Index")
        self.dockIndex.setObjectName("Index")
        self.dockSearch = QDockWidget("Search")
        self.dockSearch.setObjectName("Search")
        self.dockToc = QDockWidget("TOC")
        self.dockToc.setObjectName("TOC")
        self.dockIndex.setWidget(self.notesTree)
        self.dockSearch.setWidget(self.searchTab)
        self.dockToc.setWidget(self.tocTree)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dockIndex)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dockSearch)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dockToc)
        self.tabifyDockWidget(self.dockIndex, self.dockSearch)
        self.tabifyDockWidget(self.dockSearch, self.dockToc)
        self.setTabPosition(Qt.LeftDockWidgetArea, QTabWidget.North)
        self.dockIndex.raise_()      # Put dockIndex on top of the tab stack
        # left pane
        #self.tabWidget.addTab(self.notesTree, 'Index')
        #self.tabWidget.addTab(self.searchTab, 'Search')
        #self.tabWidget.addTab(self.tocTree, 'TOC')
        #self.tabWidget.setMinimumWidth(200)
        # self.rightSplitter.setSizes([600,20,600,580])
        self.rightSplitter.setStretchFactor(0, 0)

        # Global Actions
        actTabIndex = self.act(self.tr('Switch to Index Tab'),
                               QKeySequence('Ctrl+Shift+I'),
                               lambda: self.tabWidget.setCurrentWidget(self.notesTree))
        actTabSearch = self.act(self.tr('Switch to Search Tab'),
                                QKeySequence('Ctrl+Shift+F'),
                                lambda: self.currentTabChanged(1))
                                # lambda:self.tabWidget.setCurrentWidget(self.searchTab)
        self.addAction(actTabIndex)
        self.addAction(actTabSearch)
        
        # Shortcuts to switch notes.
        actNote1 = self.act(self.tr(""), QKeySequence("Ctrl+1"),
                            lambda: self.switchNote(1))
        actNote2 = self.act(self.tr(""), QKeySequence("Ctrl+2"),
                            lambda: self.switchNote(2))
        actNote3 = self.act(self.tr(""), QKeySequence("Ctrl+3"),
                            lambda: self.switchNote(3))
        actNote4 = self.act(self.tr(""), QKeySequence("Ctrl+4"),
                            lambda: self.switchNote(4))
        actNote5 = self.act(self.tr(""), QKeySequence("Ctrl+5"),
                            lambda: self.switchNote(5))
        actNote6 = self.act(self.tr(""), QKeySequence("Ctrl+6"),
                            lambda: self.switchNote(6))
        actNote7 = self.act(self.tr(""), QKeySequence("Ctrl+7"),
                            lambda: self.switchNote(7))
        actNote8 = self.act(self.tr(""), QKeySequence("Ctrl+8"),
                            lambda: self.switchNote(8))
        actNote9 = self.act(self.tr(""), QKeySequence("Ctrl+9"),
                            lambda: self.switchNote(9))
        self.addAction(actNote1)
        self.addAction(actNote2)
        self.addAction(actNote3)
        self.addAction(actNote4)
        self.addAction(actNote5)
        self.addAction(actNote6)
        self.addAction(actNote7)
        self.addAction(actNote8)
        self.addAction(actNote9)

        # actions in menuFile
        self.actionNewPage = self.act(
            self.tr('&New Page...'), QKeySequence.New, trig=self.notesTree.newPage)
        self.actionNewSubpage = self.act(self.tr('New Sub&page...'), QKeySequence(
            'Ctrl+Shift+N'), trig=self.notesTree.newSubpage)
        self.actionImportPage = self.act(
            self.tr('&Import Page...'), trig=self.importPage)
        self.actionOpenNotebook = self.act(
            self.tr('&Open Notebook...'), QKeySequence.Open, trig=self.openNotebook)
        self.actionSave = self.act(self.tr(
            '&Save'), QKeySequence.Save, trig=self.saveCurrentNote)
        self.actionSave.setEnabled(False)
        self.actionSaveAs = self.act(self.tr('Save &As...'), QKeySequence(
            'Ctrl+Shift+S'), trig=self.saveNoteAs)
        self.actionHtml = self.act(
            self.tr('to &HTML'), trig=self.notesEdit.saveAsHtml)
        self.actionPrint = self.act(self.tr(
            '&Print'), QKeySequence('Ctrl+P'), trig=self.printNote)
        self.actionRenamePage = self.act(self.tr(
            '&Rename Page...'), QKeySequence('F2'), trig=self.notesTree.renamePageWrapper)
        self.actionDelPage = self.act(self.tr(
            '&Delete Page'), QKeySequence('Delete'), trig=self.notesTree.delPageWrapper)
        self.actionQuit = self.act(self.tr('&Quit'), QKeySequence.Quit, SLOT('close()'))
        self.actionQuit.setMenuRole(QAction.QuitRole)
        # actions in menuEdit
        self.actionUndo = self.act(self.tr('&Undo'), QKeySequence.Undo,
                                   trig=lambda: self.notesEdit.undo())
        self.actionUndo.setEnabled(False)
        self.notesEdit.undoAvailable.connect(self.actionUndo.setEnabled)
        self.actionRedo = self.act(self.tr('&Redo'), QKeySequence.Redo,
                                   trig=lambda: self.notesEdit.redo())
        self.actionRedo.setEnabled(False)
        self.notesEdit.redoAvailable.connect(self.actionRedo.setEnabled)
        self.actionFindText = self.act(self.tr('&Find Text'), QKeySequence.Find,
            self.findBar.setVisible, True)
        self.actionFind = self.act(
            self.tr('Next'), QKeySequence.FindNext, trig=self.findText)
        self.actionFindPrev = self.act(
            self.tr('Previous'), QKeySequence.FindPrevious,
            trig=lambda: self.findText(back=True))
        self.actionSortLines = self.act(
            self.tr('&Sort Lines'), trig=self.sortLines)
        self.actionInsertImage = self.act(
            self.tr('&Insert Image'), QKeySequence('Ctrl+I'), trig=self.insertImage)
        self.actionInsertImage.setEnabled(False)

        # actions in menuView
        self.actionEdit = self.act(self.tr('Edit'), QKeySequence('Ctrl+E'),
            self.edit, True, ":/icons/edit.svg", "Edit mode (Ctrl+E)")
        self.actionSplit = self.act(self.tr('Split'), QKeySequence('Ctrl+R'),
            self.liveView, True, ":/icons/split.svg", "Split mode (Ctrl+R)")
        self.actionFlipEditAndView = self.act(
            self.tr('Flip Edit and View'), trig=self.flipEditAndView)
        self.actionFlipEditAndView.setEnabled(False)
        self.actionLeftAndRight = self.act(
            self.tr('Split into Left and Right'), trig=self.leftAndRight)
        self.actionUpAndDown = self.act(
            self.tr('Split into Up and Down'), trig=self.upAndDown)
        self.actionToggleIndex = self.act(
            self.tr('Index tab'), None, self.toggleIndex, True)
        self.actionToggleSearch = self.act(
            self.tr('Search tab'), None, self.toggleSearch, True)
        self.actionToggleToc = self.act(
            self.tr('TOC tab'), None, self.toggleToc, True)
        # self.actionLeftAndRight.setEnabled(False)
        # self.actionUpAndDown.setEnabled(False)
        # actions in menuHelp
        self.actionReadme = self.act(self.tr('README'), trig=self.readmeHelp)

        self.menuBar = QMenuBar(self)
        self.setMenuBar(self.menuBar)
        self.menuFile = self.menuBar.addMenu(self.tr('&File'))
        self.menuEdit = self.menuBar.addMenu(self.tr('&Edit'))
        self.menuView = self.menuBar.addMenu(self.tr('&View'))
        self.menuHelp = self.menuBar.addMenu(self.tr('&Help'))
        # menuFile
        self.menuFile.addAction(self.actionNewPage)
        self.menuFile.addAction(self.actionNewSubpage)
        self.menuFile.addAction(self.actionImportPage)
        self.menuFile.addAction(self.actionOpenNotebook)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addAction(self.actionSaveAs)
        self.menuFile.addAction(self.actionPrint)
        self.menuExport = self.menuFile.addMenu(self.tr('&Export'))
        self.menuExport.addAction(self.actionHtml)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionRenamePage)
        self.menuFile.addAction(self.actionDelPage)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        # menuEdit
        self.menuEdit.addAction(self.actionUndo)
        self.menuEdit.addAction(self.actionRedo)
        self.menuEdit.addAction(self.actionFindText)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionSortLines)
        self.menuEdit.addAction(self.actionInsertImage)
        # menuView
        self.menuView.addAction(self.actionEdit)
        self.menuView.addAction(self.actionSplit)
        self.menuView.addAction(self.actionFlipEditAndView)
        self.menuMode = self.menuView.addMenu(self.tr('Mode'))
        self.menuMode.addAction(self.actionLeftAndRight)
        self.menuMode.addAction(self.actionUpAndDown)
        self.menuPanel = self.menuView.addMenu(self.tr('Side Panel'))
        self.menuPanel.addAction(self.actionToggleIndex)
        self.menuPanel.addAction(self.actionToggleSearch)
        self.menuPanel.addAction(self.actionToggleToc)
        # menuHelp
        self.menuHelp.addAction(self.actionReadme)

        self.toolBar = QToolBar(self.tr("toolbar"), self)
        self.toolBar.setObjectName("toolbar")       # needed in saveState()
        self.toolBar.setIconSize(QSize(16, 16))
        self.toolBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.addToolBar(Qt.TopToolBarArea, self.toolBar)
        self.toolBar.addAction(self.actionEdit)
        self.toolBar.addAction(self.actionSplit)
        self.findEdit = QLineEdit(self.findBar)
        self.findEdit.returnPressed.connect(self.findText)
        self.checkBox = QCheckBox(self.tr('Match case'), self.findBar)
        self.findBar.addWidget(self.findEdit)
        self.findBar.addWidget(self.checkBox)
        self.findBar.addAction(self.actionFindPrev)
        self.findBar.addAction(self.actionFind)
        self.findBar.setVisible(False)
        self.findBar.visibilityChanged.connect(self.findBarVisibilityChanged)

        self.statusBar = QStatusBar(self)
        self.setStatusBar(self.statusBar)
        self.statusLabel = QLabel(self)
        self.statusBar.addWidget(self.statusLabel, 1)

        self.tabWidget.currentChanged.connect(self.currentTabChanged)

        self.notesTree.currentItemChanged.connect(
            self.currentItemChangedWrapper)
        self.tocTree.currentItemChanged.connect(self.tocNavigate)
        self.connect(self.notesEdit, SIGNAL('textChanged()'), self.noteEditted)

        self.notesEdit.document(
        ).modificationChanged.connect(self.modificationChanged)

        self.updateRecentViewedNotes()
        notes = self.settings.recentViewedNotes()
        if len(notes) != 0:
            item = self.notesTree.pagePathToItem(notes[0])
            self.notesTree.setCurrentItem(item)

    def restore(self):
        """ Restore saved geometry and state.
            Set the status of side panels in View Menu correspondently.
        """
        if self.settings.geometry:
            self.restoreGeometry(self.settings.geometry)
        if self.settings.windowstate:
            self.restoreState(self.settings.windowstate)
        self.actionToggleIndex.setChecked(
            self.dockIndex.isVisible())
        self.actionToggleSearch.setChecked(
            self.dockSearch.isVisible())
        self.actionToggleToc.setChecked(
            self.dockToc.isVisible())

    def initTree(self, notebookPath, parent):
        ''' When there exist foo.md, foo.mkd, foo.markdown, 
            only one item will be shown in notesTree.
        '''
        if not QDir(notebookPath).exists():
            return
        notebookDir = QDir(notebookPath)
        notesList = notebookDir.entryInfoList(['*.md', '*.mkd', '*.markdown'],
                                               QDir.NoFilter,
                                               QDir.Name|QDir.IgnoreCase)
        nl = [note.completeBaseName() for note in notesList]
        noduplicate = list(set(nl))
        for name in noduplicate:
            item = QTreeWidgetItem(parent, [name])
            path = notebookPath + '/' + name 
            self.initTree(path, item)

    def updateToc(self):
        ''' TOC is updated in `updateView`
            tocTree fields: [hdrText, hdrPosition, hdrAnchor]
        '''
        root = self.notesTree.currentItemName()
        self.tocTree.clear()
        item = QTreeWidgetItem(self.tocTree, [root, '0'])
        curLevel = 0
        for (l, h, p, a) in parseHeaders(self.notesEdit.toPlainText()):
            ac = len(l)             # hdrLevel is the number of asterisk
            # print("curLevel: %d, ac: %d" % (curLevel, ac))
            val = [h, str(p), a]
            if ac == curLevel:
                item = QTreeWidgetItem(item.parent(), val)
            elif ac < curLevel:
                item = QTreeWidgetItem(item.parent().parent(), val)
                curLevel = ac
            else:
                item = QTreeWidgetItem(item, val)
                curLevel = ac
        self.tocTree.expandAll()

    def openNote(self, noteFullName):
        filename = self.notesTree.noteFullPath(noteFullName)
        print(filename)
        fh = QFile(filename)
        try:
            if not fh.open(QIODevice.ReadOnly):
                raise IOError(fh.errorString())
        except IOError as e:
            QMessageBox.warning(self, 'Read Error',
                                'Failed to open %s: %s' % (filename, e))
        finally:
            if fh is not None:
                noteBody = QTextStream(fh).readAll()
                fh.close()
                self.notesEdit.setPlainText(noteBody)
                self.notesView.scrollPosition = QPoint(0, 0)
                # self.actionSave.setEnabled(False)
                self.notesEdit.document().setModified(False)
                self.notesView.updateView()
                self.setCurrentNote()
                self.updateRecentViewedNotes()
                self.viewedListActions[0].setChecked(True)
                self.statusLabel.setText(noteFullName)

    def currentTabChanged(self, index):
        self.tabWidget.setCurrentIndex(index)
        if index == 1:
            self.searchEdit.setFocus()
            self.searchEdit.selectAll()

    def currentItemChangedWrapper(self, current, previous):
        if current is None:
            return
        #if previous != None and self.notesTree.exists(previous):
        if self.notesTree.exists(self.notesTree.itemToPagePath(previous)):
            self.saveNote(previous)
        name = self.notesTree.itemToPagePath(current)
        self.openNote(name)
        # name = self.notesTree.currentItemName()

    def tocNavigate(self, current, previous):
        ''' works for notesEdit now '''
        if current is None:
            return
        pos = int(current.text(1))
        link = "file://" + self.notebookPath + "/#" + current.text(2)
        # Move cursor to END first will ensure
        # header is positioned at the top of visual area.
        self.notesEdit.moveCursor(QTextCursor.End)
        cur = self.notesEdit.textCursor()
        cur.setPosition(pos, QTextCursor.MoveAnchor)
        self.notesEdit.setTextCursor(cur)
        self.notesView.load(QUrl(link))

    def switchNote(self, num):
        self.viewedListActions[num].trigger()


    def saveCurrentNote(self):
        item = self.notesTree.currentItem()
        self.saveNote(item)

    def saveNote(self, item):
        if self.notesEdit.document().isModified():
            self.notesEdit.document().setModified(False)
        else:
            return
        pageName = item.text(0)
        filePath = os.path.join(self.notebookPath,
                                self.notesTree.itemToPagePath(item) + self.settings.fileExt)
        self.notesEdit.save(pageName, filePath)

    def saveNoteAs(self):
        self.saveCurrentNote()
        fileName = QFileDialog.getSaveFileName(self, self.tr('Save as'), '',
                                               '(*.md *.mkd *.markdown);;'+self.tr('All files(*)'))
        if fileName == '':
            return
        if not QFileInfo(fileName).suffix():
            fileName += '.md'
        fh = QFile(fileName)
        fh.open(QIODevice.WriteOnly)
        savestream = QTextStream(fh)
        savestream << self.notesEdit.toPlainText()
        fh.close()

    def printNote(self):
        printer = QPrinter(QPrinter.HighResolution)
        printer.setCreator(__appname__ + ' ' + __version__)
        printer.setDocName(self.notesTree.currentItem().text(0))
        printdialog = QPrintDialog(printer, self)
        if printdialog.exec() == QDialog.Accepted:
            self.notesView.print_(printer)

    def noteEditted(self):
        """ Continuously get fired while editing"""
        self.updateToc()
        self.notesView.updateLiveView()

    def modificationChanged(self, changed):
        """ Fired one time: modified or not """
        self.actionSave.setEnabled(changed)
        name = self.notesTree.currentItemName()
        self.statusBar.clearMessage()
        if changed:
            self.statusLabel.setText(name + '*')
        else:
            self.statusLabel.setText(name)

    def importPage(self):
        filename = QFileDialog.getOpenFileName(
            self, self.tr('Import file'), '',
            '(*.md *.mkd *.markdown *.txt);;'+self.tr('All files(*)'))
        if filename == '':
            return
        self.importPageCore(filename)

    def importPageCore(self, filename):
        fh = QFile(filename)
        fh.open(QIODevice.ReadOnly)
        fileBody = QTextStream(fh).readAll()
        fh.close()
        note = QFileInfo(filename)
        path = os.path.join(self.notebookPath, 
                            note.completeBaseName() + self.settings.fileExt)
        fh = QFile(path)
        if fh.exists():
            QMessageBox.warning(self, 'Import Error',
                                'Page already exists: %s' % note.completeBaseName())
            return
        fh.open(QIODevice.WriteOnly)
        savestream = QTextStream(fh)
        savestream << fileBody
        fh.close()
        QTreeWidgetItem(self.notesTree, [note.completeBaseName()])
        self.notesTree.sortItems(0, Qt.AscendingOrder)
        item = self.notesTree.pagePathToItem(note.completeBaseName())
        self.notesTree.setCurrentItem(item)

    def openNotebook(self):
        dialog = NotebookListDialog(self)
        if dialog.exec_():
            pass

    def act(self, name, shortcut=None, trig=None, checkable=False, 
            icon=None, tooltip=None):
        """ A wrapper to several QAction methods """
        if icon:
            action = QAction(QIcon(icon), name, self)
        else:
            action = QAction(name, self)
        if shortcut:
            action.setShortcut(shortcut)
        action.setCheckable(checkable)
        if tooltip:
            action.setToolTip(tooltip)
        self.connect(action, SIGNAL('triggered(bool)'), trig)
        return action

    def edit(self, viewmode):
        """ Switch between EDIT and VIEW mode. """

        if self.actionSplit.isChecked():
            self.actionSplit.setChecked(False)
        self.notesView.setVisible(not viewmode)
        self.notesEdit.setVisible(viewmode)
        
        # Gives the keyboard input focus to notesEdit/notesView.
        # Without this, keyboard input may change note text even when
        # notesEdit is invisible. 
        if viewmode:
            self.notesEdit.setFocus()
        else:
            self.notesView.setFocus()
        
        self.saveCurrentNote()
        self.actionInsertImage.setEnabled(viewmode)
        self.actionLeftAndRight.setEnabled(True)
        self.actionUpAndDown.setEnabled(True)

        # Render the note text as it is.
        self.notesView.updateView()

    def liveView(self, viewmode):
        """ Switch between VIEW and LIVE VIEW mode. """

        self.actionSplit.setChecked(viewmode)
        sizes = self.noteSplitter.sizes()
        if self.actionEdit.isChecked():
            self.actionEdit.setChecked(False)
            self.notesView.setVisible(viewmode)
            splitSize = [sizes[0]*0.45, sizes[0]*0.55]
        else:
            self.notesEdit.setVisible(viewmode)
            splitSize = [sizes[1]*0.45, sizes[1]*0.55]
        
        # setFocus for the same reason as in edit(self, viewmode)
        if viewmode:
            self.notesEdit.setFocus()
        else:
            self.notesView.setFocus()
        
        self.actionFlipEditAndView.setEnabled(viewmode)
        self.actionUpAndDown.setEnabled(viewmode)
        self.actionInsertImage.setEnabled(viewmode)
        self.noteSplitter.setSizes(splitSize)
        self.saveCurrentNote()

        # Render the note text as it is.
        self.notesView.updateView()

    def findBarVisibilityChanged(self, visible):
        self.actionFindText.setChecked(visible)
        if visible:
            self.findEdit.setFocus(Qt.ShortcutFocusReason)

    def findText(self, back=False):
        flags = 0
        if back:
            flags = QTextDocument.FindBackward
        if self.checkBox.isChecked():
            flags = flags | QTextDocument.FindCaseSensitively
        text = self.findEdit.text()
        if not self.findMain(text, flags):
            if text in self.notesEdit.toPlainText():
                cursor = self.notesEdit.textCursor()
                if back:
                    cursor.movePosition(QTextCursor.End)
                else:
                    cursor.movePosition(QTextCursor.Start)
                self.notesEdit.setTextCursor(cursor)
                self.findMain(text, flags)
        # self.notesView.findText(text, flags)

    def findMain(self, text, flags):
        viewFlags = QWebPage.FindFlags(
            flags) | QWebPage.FindWrapsAroundDocument
        if flags:
            self.notesView.findText(text, viewFlags)
            return self.notesEdit.find(text, flags)
        else:
            self.notesView.findText(text)
            return self.notesEdit.find(text)

    def sortLines(self):
        ''' sort selected lines
            Currently, have to select whole lines. (ToFix)
            TODO: second sort reverse the order
        '''
        cursor = self.notesEdit.textCursor()
        text = cursor.selectedText()
        lines = text.split('\u2029')      # '\u2029' is the line break
        sortedLines = sorted(lines)
        self.notesEdit.insertPlainText('\n'.join(sortedLines))

    def insertImage(self):
        # TODO how to include all image types?
        filename = QFileDialog.getOpenFileName(
            self, self.tr('Insert image'), '',
            '(*.jpg *.png *.gif *.tif);;'+self.tr('All files(*)'))
        filename = '![](file://' + filename + ')'
        self.notesEdit.insertPlainText(filename)

    def notesEditInFocus(self, e):
        print('hello')
        if e.gotFocus:
            self.actionInsertImage.setEnabled(True)
        # if e.lostFocus:
        #    self.actionInsertImage.setEnabled(False)

        # QWidget.focusInEvent(self,f)

    def searchNote(self):
        pattern = self.searchEdit.text()
        qres = []
        with self.ix.searcher() as searcher:
            queryp = QueryParser("content", self.ix.schema)
            queryp.add_plugin(RegexPlugin())
            query = queryp.parse('r"' + pattern + '"')
                                 # r"pattern" is the desired regex term format
            pathFacet = sorting.FieldFacet("path")
            scores = sorting.ScoreFacet()
            results = searcher.search(
                query, limit=None, sortedby=[pathFacet, scores])  # default limit is 10!
            for r in results:
                listItem = QListWidgetItem()
                title = r['title']
                text = r['path']
                term = r.highlights("content")
                qres.append([title, text, term])
            html = """
                    <style>
                        body { font-size: 14px; }
                        .path { font-size: 12px; color: #009933; }
                    </style>
                   """
            for ti, te, hi in qres:
                html += ("<p><a href='" + te + "'>" + ti + 
                         "</a><br/><span class='path'>" + 
                        te + "</span><br/>" + hi + "</p>")
            self.searchView.setHtml(html)

    def whoosh_index(self):
        it = QTreeWidgetItemIterator(
            self.notesTree, QTreeWidgetItemIterator.All)
        writer = self.ix.writer()
        while it.value():
            treeItem = it.value()
            name = self.notesTree.itemToPagePath(treeItem)
            path = os.path.join(self.notesTree.noteFullPath(name))
            print(path)
            fileobj = open(path, 'r')
            content = fileobj.read()
            fileobj.close()
            writer.add_document(
                path=name, title=parseTitle(content, name), content=content)
            it += 1
        writer.commit()

    def listItemChanged(self, row):
        if row != -1:
            item = self.searchList.currentItem().data(Qt.UserRole)
            self.notesTree.setCurrentItem(item)
            flags = QWebPage.HighlightAllOccurrences
            self.notesView.findText(self.searchEdit.text(), flags)

    def setCurrentNote(self):
        item = self.notesTree.currentItem()
        name = self.notesTree.itemToPagePath(item)

        # Current note is inserted to head of list.
        notes = self.settings.recentViewedNotes()
        for f in notes:
            if f == name:
                notes.remove(f)
        notes.insert(0, name)

        # TODO: move this NUM to configuration
        if len(notes) > 20:
            del notes[20:]
        self.settings.updateRecentViewedNotes(notes)

    def updateRecentViewedNotes(self):
        """ Switching notes will triger this. """

        self.viewedList.clear()
        self.viewedListActions = []

        # Check notes exists.
        viewedNotes = self.settings.recentViewedNotes()
        existedNotes = []
        for f in viewedNotes:
            if self.notesTree.exists(f):
                existedNotes.append(f)
                splitName = f.split('/')
                self.viewedListActions.append(
                    self.act(splitName[-1], None, self.openFunction(f), True))
                
        self.settings.updateRecentViewedNotes(existedNotes)
        for action in self.viewedListActions:
            self.viewedList.addAction(action)

    def openFunction(self, name):
        item = self.notesTree.pagePathToItem(name)
        return lambda: self.notesTree.setCurrentItem(item)

    def toggleIndex(self, visible):
        self.actionToggleIndex.setChecked(visible)
        self.toggleTab(self.dockIndex, visible)

    def toggleSearch(self, visible):
        self.actionToggleSearch.setChecked(visible)
        self.toggleTab(self.dockSearch, visible)

    def toggleToc(self, visible):
        self.actionToggleToc.setChecked(visible)
        self.toggleTab(self.dockToc, visible)

    def toggleTab(self, tab, visible):
        if tab.isVisible():
            tab.hide()
        else:
            tab.show()

    def flipEditAndView(self):
        index = self.noteSplitter.indexOf(self.notesEdit)
        if index == 0:
            self.noteSplitter.insertWidget(1, self.notesEdit)
        else:
            self.noteSplitter.insertWidget(0, self.notesEdit)

    def leftAndRight(self):
        self.liveView(True)
        self.noteSplitter.setOrientation(Qt.Horizontal)
        self.actionLeftAndRight.setEnabled(False)
        self.actionUpAndDown.setEnabled(True)

    def upAndDown(self):
        self.liveView(True)
        self.noteSplitter.setOrientation(Qt.Vertical)
        self.actionUpAndDown.setEnabled(False)
        self.actionLeftAndRight.setEnabled(True)

    def readmeHelp(self):
        readmeFile = '/usr/share/mikidown/README.mkd'
        self.importPageCore(readmeFile)

    def closeEvent(self, event):
        """
            saveGeometry: Saves the current geometry and state for 
                          top-level widgets
            saveState: Restores the state of this mainwindow's toolbars 
                       and dockwidgets 
        """
        self.saveCurrentNote()
        self.settings.saveGeometry(self.saveGeometry())
        self.settings.saveWindowState(self.saveState())
        #self.settings.setValue("geometry", self.saveGeometry())
        #self.settings.setValue("windowstate", self.saveState())
        event.accept()



import re

from PyQt4.QtCore import QDir, QUrl, QPoint, QTimer
from PyQt4.QtWebKit import QWebView, QWebPage
import markdown


class MikiView(QWebView):

    def __init__(self, parent=None):
        super(MikiView, self).__init__(parent)
        self.parent = parent

        self.settings().clearMemoryCaches()
        self.notebookPath = parent.settings.notebookPath
        self.settings().setUserStyleSheetUrl(
            QUrl.fromLocalFile(self.notebookPath + '/notes.css'))
        self.page().setLinkDelegationPolicy(QWebPage.DelegateAllLinks)

        self.page().linkClicked.connect(self.linkClicked)
        self.page().linkHovered.connect(self.linkHovered)
        self.page().mainFrame(
        ).contentsSizeChanged.connect(self.contentsSizeChanged)

        self.scrollPosition = QPoint(0, 0)

    def linkClicked(self, qurl):
        '''three kinds of link:
            external uri: http/https
            page ref link:
            toc anchor link: #
        '''
        name = qurl.toString()
        print(name)
        http = re.compile('https?://')
        if http.match(name):                        # external uri
            QDesktopServices.openUrl(qurl)
            return

        self.load(qurl)
        name = name.replace('file://', '')
        name = name.replace(self.notebookPath, '').split('#')
        item = self.parent.notesTree.pagePathToItem(name[0])
        if not item or item == self.parent.notesTree.currentItem():
            return
        else:
            self.parent.notesTree.setCurrentItem(item)
            if len(name) > 1:
                link = "file://" + self.notebookPath + "/#" + name[1]
                self.load(QUrl(link))
            viewFrame = self.page().mainFrame()
            self.scrollPosition = viewFrame.scrollPosition()

    def linkHovered(self, link, title, textContent):
        '''show link in status bar
            ref link shown as: /parent/child/pageName
            toc link shown as: /parent/child/pageName#anchor (ToFix)
        '''
        # TODO: link to page by: /parent/child/pageName#anchor
        if link == '':                              # not hovered
            self.parent.statusBar.showMessage(self.parent.notesTree.currentItemName())
        else:                                       # beautify link
            link = link.replace('file://', '')
            link = link.replace(self.notebookPath, '')
            self.parent.statusBar.showMessage(link)

    def contentsSizeChanged(self, newSize):
        '''scroll notesView while editing (adding new lines)
           Whithout this, every `updateView` will result in scroll to top.
        '''
        if self.scrollPosition == QPoint(0, 0):
            return
        viewFrame = self.page().mainFrame()
        newY = self.scrollPosition.y(
        ) + newSize.height() - self.contentsSize.height()
        self.scrollPosition.setY(newY)
        viewFrame.setScrollPosition(self.scrollPosition)

    def updateView(self):
        # url_notebook = 'file://' + os.path.join(self.notebookPath, '/')
        viewFrame = self.page().mainFrame()
        # Store scrollPosition before update notesView
        self.scrollPosition = viewFrame.scrollPosition()
        self.contentsSize = viewFrame.contentsSize()
        url_notebook = 'file://' + self.notebookPath + '/'
        self.setHtml(self.parent.notesEdit.toHtml(), QUrl(url_notebook))
        # Restore previous scrollPosition
        viewFrame.setScrollPosition(self.scrollPosition)

    def updateLiveView(self):
        if self.parent.actionSplit.isChecked():
            QTimer.singleShot(1000, self.updateView)




import re

from PyQt4.QtGui import QSyntaxHighlighter, QColor, QFont, QTextCharFormat


class MikiHighlighter(QSyntaxHighlighter):

    def __init__(self, parent=None):
        super(MikiHighlighter, self).__init__(parent)
        baseFontSize = 12
        NUM = 15
        self.patterns = []
        regexp = [0] * NUM
        font = [0]*NUM
        color = [0]*NUM
        # 0: html tags - <pre></pre>
        regexp[0] = '</?[^>]+>'
        font[0] = QFont("monospace", baseFontSize, -1)
        color[0] = QColor("#A40000")
        # 1: h1 - #
        regexp[1] = '^#[^#]+'
        color[1] = QColor("#4E9A06")
        font[1] = QFont("decorative", 2*baseFontSize, QFont.Bold)
        # 2: h2 - ##
        regexp[2] = '^##[^#]+'
        color[2] = QColor("#4E9A06")
        font[2] = QFont("serif", 5.0/3*baseFontSize, QFont.Bold)
        # 3: h3 - ###
        regexp[3] = '^###[^#]+'
        color[3] = QColor("#4E9A06")
        font[3] = QFont("serif", 4.0/3*baseFontSize, QFont.Bold)
        # 4: h4 and more - ####
        regexp[4] = '^####.+'
        color[4] = QColor("#4E9A06")
        font[4] = QFont("serif", baseFontSize, QFont.Bold)
        # 5: html symbols - &gt;
        regexp[5] = '&[^; ].+;'
        color[5] = QColor("#A40000")
        font[5] = QFont("monospace", baseFontSize, -1)
        # 6: html comments - <!-- -->
        regexp[6] = '<!--.+-->'
        color[6] = QColor("#888A85")
        font[6] = QFont(None, baseFontSize, -1)
        # 7: delete - ~~delete~~
        regexp[7] = '~~[^~~]*~~'
        color[7] = QColor("#888A85")
        font[7] = QFont(None, baseFontSize, -1)
        # 8: insert - __insert__
        regexp[8] = '__[^__]*__'
        font[8] = QFont(None, baseFontSize, -1)
        font[8].setUnderline(True)
        # 9: strong - **strong**
        regexp[9] = '\*\*[^**]*\*\*'
        color[9] = QColor("#F57900")
        font[9] = QFont(None, baseFontSize, QFont.Bold)
        # 10: emphasis - //emphasis//
        regexp[10] = r'//[^//\(\)]*//'
        color[10] = QColor("#F57900")
        font[10] = QFont(None, baseFontSize, -1, True)
        # 11: links - (links) after [] or links after []:
        regexp[11] = r'(?<=(\]\())[^\(\)]*(?=\))'
        font[11] = QFont(None, baseFontSize, -1, True)
        font[11].setUnderline(True)
        #.setUnderlineColor("#204A87")
        # 12: link/image references - [] or ![]
        regexp[12] = r'!?\[[^\[\]]*\]'
        color[12] = QColor("#204A87")
        font[12] = QFont(None, baseFontSize, -1)
        # 13: blockquotes and lists -  > or - or *
        regexp[13] = r'(^>+)|(^- )|(^\* )'
        color[13] = QColor("#F57900")
        font[13] = QFont(None, baseFontSize, -1)
        # 14: fence - ``` or ~~~
        regexp[14] = '^(?:~{3,}|`{3,}).*$'
        color[14] = QColor("#F57900")
        font[14] = QFont(None, baseFontSize, QFont.Bold)
        for i in range(NUM):
            p = re.compile(regexp[i])
            f = QTextCharFormat()
            if font[i] != 0:
                f.setFont(font[i])
            if color[i] != 0:
                f.setForeground(color[i])
            self.patterns.append((p, f))

        fenced_font = QFont("monospace", baseFontSize, -1)
        self.fenced_block = re.compile("^(?:~{3,}|`{3,}).*$")
        self.fenced_format = QTextCharFormat()
        self.fenced_format.setFont(fenced_font)

    def highlightBlock(self, text):
        # highlight patterns
        for i in range(0, len(self.patterns)):
            p = self.patterns[i]
            for match in p[0].finditer(text):
                self.setFormat(
                    match.start(), match.end() - match.start(), p[1])

        # escape highlights in fenced_block
        m = self.fenced_block.match(text)
        self.setCurrentBlockState(0)
        if self.previousBlockState() != 1:
            if m:
                self.setCurrentBlockState(1)
        else:
            if m:
                self.setCurrentBlockState(0)
            else:
                self.setCurrentBlockState(1)
                self.setFormat(0, len(text), self.fenced_format)



import os
import re
import sys

from PyQt4.QtCore import QSettings
from PyQt4.QtGui import QApplication, QIcon

import mikidown.mikidown_rc
from mikidown.config import *
from mikidown.mikitray import MikiTray
from mikidown.mikiwindow import MikiWindow
from mikidown.mikibook import Mikibook

sys.path.append(os.path.dirname(__file__))


def main():

    # Instantiate a QApplication first.
    # Otherwise, Mikibook.create() won't function.
    app = QApplication(sys.argv)

    # Read notebookList, open the first notebook.
    notebooks = Mikibook.read()
    if len(notebooks) == 0:
        Mikibook.create()
        notebooks = Mikibook.read()

    settings = Setting(notebooks)
    # Initialize application and main window.
    icon = QIcon(":/icons/mikidown.svg")
    app.setWindowIcon(icon)
    window = MikiWindow(settings)
    window.show()
    window.restore()        # Restore after window show.
    tray = MikiTray(icon, window)
    tray.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



"""
HeaderLink Extension for Python-Markdown
========================================

!!HeaderID extension must be enabled to use this!!

With HeaderId, #Header will become

    <h1 id="header">Header</h1>

With HeaderLink, #Header will become

    <h1 id="header">Header<a class="headerlink" href="#header">¶</a></h1>
"""

from markdown import Extension
from markdown.treeprocessors import Treeprocessor
from markdown.util import etree


class HeaderLinkTreeprocessor(Treeprocessor):
    """ Insert anchors to headers. """

    def run(self, doc):
        for elem in doc.getiterator():
            if elem.tag in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
                if "id" in elem.attrib:
                    id = elem.get('id')
                    a = etree.SubElement(elem, "a")
                    a.set('class', 'headerlink')
                    a.set('href', '#' + id)
                    a.text = '¶'


class HeaderLinkExtension(Extension):

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = HeaderLinkTreeprocessor()
        self.processor.md = md
        md.treeprocessors.add('headerlink', self.processor, '>headerid')


def makeExtension(configs=None):
    return HeaderLinkExtension(configs=configs)



import re
from markdown.extensions.headerid import slugify, unique


def parseHeaders(source):
    ''' Quite basic header parser
        Headers are used to construct Table Of Contents

        return: [(hdrLevel, hdrText, hdrPosition, hdrAnchor)]
    '''
    # RE = re.compile(r'(^|\n)(?P<level>#{1,6})(?P<header>.*?)#*(\n|$)')
    hdrs = []
    used_ids = set()           # In case there are headers with the same name.
    RE = re.compile(r'^(#+)(.+)', re.MULTILINE)
    for m in RE.finditer(source):
        hdrLevel = m.group(1)
        hdr = m.group(2)
        pos = m.start()
        anchor = unique(slugify(hdr, '-'), used_ids)
        hdrs.append((hdrLevel, hdr, pos, anchor))
    return hdrs

def parseTitle(source, fallback):
    """ Quite basic title parser, the first header1 is taken as title """
    title_re = re.compile(r'^#([^#].+)')
    title = title_re.search(source)
    if title:
        return title.group(1).strip()
    else:
        return fallback



import os

from PyQt4.QtCore import QSettings
from whoosh import fields


__appname__ = 'mikidown'
__version__ = '0.2.2'

class Setting():

    def __init__(self, notebooks):

        # Index directory of whoosh, located in notebookPath.
        self.schema = fields.Schema(
            path = fields.ID(stored=True, unique=True, spelling=True), 
            title = fields.KEYWORD(stored=True, scorable=True),
            content = fields.TEXT(stored=True))
        self.indexdir = ".indexdir"

        self.notebookName = notebooks[0][0]
        self.notebookPath = notebooks[0][1]
        self.configfile = os.path.join(self.notebookPath, "notebook.conf")
        
        self.qsettings = QSettings(self.configfile, QSettings.NativeFormat)
        self.geometry = self.qsettings.value("geometry")
        self.windowstate = self.qsettings.value("windowstate")

        if os.path.exists(self.configfile):
            self.extensions = readListFromSettings(self.qsettings,
                                                      "extensions")
            self.fileExt = self.qsettings.value("fileExt")
        else:
            self.extensions = []
            self.fileExt = ".md"
            self.qsettings.setValue("fileExt", self.fileExt)

        if not self.extensions:
            # Default enabled python-markdown extensions.
            # http://pythonhosted.org/Markdown/extensions/index.html
            self.extensions = [
                   'nl2br'           # newline to break
                 , 'strkundr'        # bold-italics-underline-delete style
                 , 'codehilite'      # code syntax highlight
                 , 'fenced_code'     # code block
                 , 'headerid'        # add id to headers
                 , 'headerlink'      # add anchor to headers
                 , 'footnotes'
                 ]
            writeListToSettings(self.qsettings, "extensions", self.extensions)

    def saveGeometry(self, geometry):
        self.qsettings.setValue("geometry", geometry)

    def saveWindowState(self, state):
        self.qsettings.setValue("windowstate", state)

    def recentViewedNotes(self):
        return readListFromSettings(self.qsettings, "recentViewedNoteList")

    def updateRecentViewedNotes(self, notesList):
        writeListToSettings(self.qsettings, "recentViewedNoteList", notesList)

def readListFromSettings(settings, key):
    if not settings.contains(key):
        return []
    value = settings.value(key)
    if isinstance(value, str):
        return [value]
    else:
        return value


def writeListToSettings(settings, key, value):
    if len(value) >= 1:
        settings.setValue(key, value)
    else:
        settings.remove(key)



"""
Notebook management module.
"""

import os

from PyQt4.QtCore import *
from PyQt4.QtGui import *

import mikidown
from mikidown.config import *


class ListDelegate(QAbstractItemDelegate):
    """ Customize view and behavior of notebook list """

    def __init__(self, parent=None):
        super(ListDelegate, self).__init__(parent)

    def paint(self, painter, option, index):
        r = option.rect
        fontPen = QPen(QColor.fromRgb(51, 51, 51), 1, Qt.SolidLine)

        if option.state & QStyle.State_Selected:
            painter.setBrush(Qt.cyan)
            painter.drawRect(r)
        else:
            painter.setBrush(
                Qt.white if (index.row() % 2) == 0 else QColor(252, 252, 252))
            painter.drawRect(r)

        painter.setPen(fontPen)

        name = index.data(Qt.DisplayRole)
        path = index.data(Qt.UserRole)

        imageSpace = 10
        # notebook name
        r = option.rect.adjusted(imageSpace, 0, -10, -20)
        painter.setFont(QFont('Lucida Grande', 10, QFont.Bold))
        painter.drawText(r.left(), r.top(
        ), r.width(), r.height(), Qt.AlignBottom|Qt.AlignLeft, name)
        # notebook path
        r = option.rect.adjusted(imageSpace, 20, -10, 0)
        painter.setFont(QFont('Lucida Grande', 8, QFont.Normal))
        painter.drawText(
            r.left(), r.top(), r.width(), r.height(), Qt.AlignLeft, path)

    def sizeHint(self, option, index):
        return QSize(200, 40)


class NotebookListDialog(QDialog):
    """ Funtions to display, create, remove, modify notebookList """

    def __init__(self, parent=None):
        super(NotebookListDialog, self).__init__(parent)
        
        self.notebookList = QListWidget()
        self.moveUp = QPushButton('<<')
        self.moveDown = QPushButton('>>')
        self.add = QPushButton('Add')
        self.remove = QPushButton('Remove')
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                          QDialogButtonBox.Cancel)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        layout = QGridLayout()
        layout.addWidget(self.notebookList, 0, 0, 4, 6)
        layout.addWidget(self.moveUp, 1, 6)
        layout.addWidget(self.moveDown, 2, 6)
        layout.addWidget(self.add, 4, 0)
        layout.addWidget(self.remove, 4, 1)
        layout.addWidget(self.buttonBox, 4, 5, 1, 2)
        self.setLayout(layout)

        self.notebookList.setItemDelegate(ListDelegate(self.notebookList))

        self.notebookList.currentRowChanged.connect(self.updateUi)
        self.add.clicked.connect(self.actionAdd)
        self.remove.clicked.connect(self.actionRemove)
        self.moveUp.clicked.connect(self.moveItemUp)
        self.moveDown.clicked.connect(self.moveItemDown)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.initList()

    def initList(self):
        self.notebookList.clear()
        notebooks = Mikibook.read()
        for nb in notebooks:
            item = QListWidgetItem()
            item.setData(Qt.DisplayRole, nb[0])
            item.setData(Qt.UserRole, nb[1])
            self.notebookList.addItem(item)

        self.updateUi(len(notebooks) != 0)
        self.notebookList.setCurrentRow(0)
            # QListWidgetItem(nb, self.notebookList)

    def updateUi(self, row):
        flag = (row != -1)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(flag)
        self.remove.setEnabled(flag)
        self.moveUp.setEnabled(flag)
        self.moveDown.setEnabled(flag)

    def actionAdd(self):
        Mikibook.create()
        self.initList()
        count = self.notebookList.count()
        self.notebookList.setCurrentRow(count-1)

    def actionRemove(self):
        item = self.notebookList.currentItem()
        row = self.notebookList.currentRow()
        name = item.data(Qt.DisplayRole)
        path = item.data(Qt.UserRole)
        self.notebookList.takeItem(row)

        Mikibook.remove(name, path)

    def moveItemUp(self):
        item = self.notebookList.currentItem()
        row = self.notebookList.currentRow()
        if row != 0:
            # self.notebookList.removeItemWidget(item)
            self.notebookList.takeItem(row)
            self.notebookList.insertItem(row-1, item)
            self.notebookList.setCurrentRow(row-1)

    def moveItemDown(self):
        item = self.notebookList.currentItem()
        row = self.notebookList.currentRow()
        count = self.notebookList.count()
        if row != count-1:
            self.notebookList.takeItem(row)
            self.notebookList.insertItem(row+1, item)
            self.notebookList.setCurrentRow(row+1)

    def accept(self):
        notebookPath = self.notebookList.currentItem().data(Qt.UserRole)
        notebookName = self.notebookList.currentItem().data(Qt.DisplayRole)
        settings = Setting([[notebookName, notebookPath]])
        window = mikidown.MikiWindow(settings)
        window.show()
        count = self.notebookList.count()
        notebooks = []
        for i in range(count):
            name = self.notebookList.item(i).data(Qt.DisplayRole)
            path = self.notebookList.item(i).data(Qt.UserRole)
            notebooks.append([name, path])
            Mikibook.write(notebooks)

        QDialog.accept(self)


class NewNotebookDlg(QDialog):
    def __init__(self, parent=None):
        super(NewNotebookDlg, self).__init__(parent)
        self.setWindowTitle('Add Notebook - mikidown')
        tipLabel = QLabel('Choose a name and folder for your notebook.' +
                          '\nThe folder can be an existing notebook folder.')
        self.nameEditor = QLineEdit()
        self.nameEditor.setText('Notes')
        nameLabel = QLabel('Name:')
        nameLabel.setBuddy(self.nameEditor)
        self.pathEditor = QLineEdit()
        # self.pathEditor.setText('~/mikidown')
        self.pathEditor.setText(os.environ['HOME']+'/mikinotes')
        pathLabel = QLabel('Path:')
        pathLabel.setBuddy(self.pathEditor)
        browse = QPushButton('Browse')
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                     QDialogButtonBox.Cancel)

        grid = QGridLayout()
        grid.setRowMinimumHeight(1, 10)
        grid.setRowMinimumHeight(4, 10)
        grid.addWidget(tipLabel, 0, 0, 1, 4)
        grid.addWidget(nameLabel, 2, 0)
        grid.addWidget(self.nameEditor, 2, 1, 1, 4)
        grid.addWidget(pathLabel, 3, 0)
        grid.addWidget(self.pathEditor, 3, 1, 1, 4)
        grid.addWidget(browse, 3, 5)
        grid.addWidget(buttonBox, 5, 4, 1, 2)
        self.setLayout(grid)

        self.connect(browse, SIGNAL("clicked()"),
                     self.browse)
        self.connect(buttonBox, SIGNAL("accepted()"), self.accept)
        self.connect(buttonBox, SIGNAL("rejected()"), self.reject)
        self.connect(self, SIGNAL('close()'), SLOT('close()'))

    def browse(self):
        default = os.environ['HOME']
        path = QFileDialog.getExistingDirectory(self,
                                                "Select Folder",
                                                default,
                                                QFileDialog.ShowDirsOnly)
        self.pathEditor.setText(path)

    def closeEvent(self, event):
        event.accept()


class Mikibook():

    # ~/.config/mikidown/mikidown.conf
    settings = QSettings('mikidown', 'mikidown')

    def read():
        """ Read notebook list from config file """
        return readListFromSettings(Mikibook.settings, 'notebookList')

    def write(notebooks):
        """ Write notebook list to config file """
        return writeListToSettings(
            Mikibook.settings, 'notebookList', notebooks)

    def create():
        """ Display a dialog to set notebookName and notebookPath """
        newNotebook = NewNotebookDlg()
        if newNotebook.exec_():
            notebookName = newNotebook.nameEditor.text()
            notebookPath = newNotebook.pathEditor.text()
            Mikibook.add(notebookName, notebookPath)

    def add(notebookName, notebookPath):
        """ Called by create()
            Initialise the notebook directory.
        """
        # Make sure there is notes.css in notebookPath
        if not os.path.isdir(notebookPath):
            os.makedirs(notebookPath)
        cssFile = os.path.join(notebookPath, 'notes.css')
        if os.path.exists('/usr/share/mikidown/notes.css'):
            cssTemplate = '/usr/share/mikidown/notes.css'
        else:
            cssTemplate = os.path.join(
                os.path.dirname(__file__), 'notes.css')
        # If //cssFile// already exists, copy() returns false!
        QFile.copy(cssTemplate, cssFile)

        # If settings==None, initialize notebook dir without writing to 
        # configuration file.
        if Mikibook.settings:
            notebooks = Mikibook.read() 
            notebooks.append([notebookName, notebookPath])
            # TODO: make mikidown.conf become plain text
            Mikibook.write(notebooks)

    def remove(name, path):
        notebooks = Mikibook.read()
        notebooks.remove([name, path])
        Mikibook.write(notebookList)



#!/usr/bin/env python3

import sys
import logging
import mikidown

# Check the python version
try:
    version_info = sys.version_info
    assert version_info > (3, 0)
except:
    print('ERROR: mikidown needs python >= 3.0', file=sys.stderr)
    sys.exit(1)

# Run mikidown
try:
    mikidown.main()
except KeyboardInterrupt:
    print('Interrupt', file=sys.stderr)
    sys.exit(1)
else:
    sys.exit(0)



import os
import datetime

from PyQt4.QtCore import *
from PyQt4.QtGui import *
from whoosh.index import open_dir
from whoosh.qparser import QueryParser

from mikidown.config import Setting


class ItemDialog(QDialog):
    def __init__(self, parent=None):
        super(ItemDialog, self).__init__(parent)
        self.editor = QLineEdit()
        editorLabel = QLabel("Page Name:")
        editorLabel.setBuddy(self.editor)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                          QDialogButtonBox.Cancel)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        layout = QGridLayout()
        layout.addWidget(editorLabel, 0, 0)
        layout.addWidget(self.editor, 0, 1)
        layout.addWidget(self.buttonBox, 1, 1)
        self.setLayout(layout)
        self.connect(self.editor, SIGNAL("textEdited(QString)"),
                     self.updateUi)
        self.connect(self.buttonBox, SIGNAL("accepted()"), self.accept)
        self.connect(self.buttonBox, SIGNAL("rejected()"), self.reject)

    def setPath(self, path):
        self.path = path

    def setText(self, text):
        self.editor.setText(text)
        self.editor.selectAll()

    def updateUi(self):
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            self.editor.text() != "")

    def accept(self):
        if self.path == '':
            notePath = self.editor.text()
        else:
            notePath = self.path + '/' + self.editor.text()

        if QFile.exists(notePath+'.md') or QFile.exists(notePath+'.mkd') or QFile.exists(notePath+'.markdown'):
            QMessageBox.warning(self, 'Error',
                                'Page already exists: %s' % notePath)
        else:
            QDialog.accept(self)


class MikiTree(QTreeWidget):

    def __init__(self, parent=None):
        super(MikiTree, self).__init__(parent)
        self.settings = parent.settings
        self.notebookPath = self.settings.notebookPath

        self.header().close()
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        # self.setDropIndicatorShown(True)
        self.setDragDropOverwriteMode(True)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        # self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setContextMenuPolicy(Qt.CustomContextMenu)

        self.customContextMenuRequested.connect(self.treeMenu)
        self.indexdir = os.path.join(self.notebookPath, self.settings.indexdir)

    def itemToPagePath(self, item):
        path = ''
        if not hasattr(item, 'text'):
            return path
        path = item.text(0)
        parent = item.parent()
        while parent is not None:
            path = parent.text(0) + '/' + path
            parent = parent.parent()
        return path

    def pagePathToItem(self, name):
        if name[0] == '/':
            name = name[1:len(name)]
        if name[-1] == '/':
            name = name[0:-1]
        splitPath = name.split('/')
        depth = len(splitPath)
        # print(depth)
        itemList = self.findItems(
            splitPath[depth-1], Qt.MatchExactly|Qt.MatchRecursive)
        if len(itemList) == 1:
            # print(itemList[0].text(0))
            return itemList[0]
        for item in itemList:
            path = self.itemToPagePath(item)
            if name == path:
                return item
            '''
            parent = item.parent()
            for i in range(depth):
                print(i)
                if parent == None:
                    if i == depth-1:
                        return item
                    else:
                        continue
                if depth-i-2 < 0:
                    break
                if parent.text(0) == splitPath[depth-i-2]:
                    if depth-i-2 == 0:
                        return item
                    else:
                        parent = parent.parent()
            '''

    def noteFullPath(self, noteFullName):
        ''' FullPath = notebookPath + noteFullName + fileExt
        fileExt is stored in notebook.conf 
        '''
        extName = ['.md', '.mkd', '.markdown']
        defExt = self.settings.fileExt
        extName.remove(defExt)
        extName.insert(0, defExt)
        for ext in extName:
            filename = os.path.join(self.notebookPath,
                                    noteFullName + ext)
            if QFile.exists(filename):
                return filename

        return ""

    def currentItemName(self):
        item = self.currentItem()
        return self.itemToPagePath(item)

    def getPath(self, item):
        path = ''
        if not hasattr(item, 'text'):
            return path
        item = item.parent()
        while item is not None:
            path = item.text(0) + '/' + path
            item = item.parent()
        return path

    def treeMenu(self):
        menu = QMenu()
        menu.addAction("New Page...", self.newPage)
        menu.addAction("New Subpage...", self.newSubpage)
        menu.addSeparator()
        menu.addAction("Collapse This Note Tree",
                       lambda item=self.currentItem(): self.recurseCollapse(item))
        menu.addAction("Uncollapse This Note Tree",
                       lambda item=self.currentItem():  self.recurseExpand(item))
        menu.addAction("Collapse All", self.collapseAll)
        menu.addAction("Uncollapse All", self.expandAll)
        menu.addSeparator()
        menu.addAction('Rename Page...',
                       lambda item=self.currentItem(): self.renamePage(item))
        self.delCallback = lambda item=self.currentItem(): self.delPage(item)
        menu.addAction("Delete Page", self.delCallback)
        menu.exec_(QCursor.pos())

    def newPage(self, name=None):
        if self.currentItem() is None:
            self.newPageCore(self, name)
        else:
            parent = self.currentItem().parent()
            if parent is not None:
                self.newPageCore(parent, name)
            else:
                self.newPageCore(self, name)

    def newSubpage(self, name=None):
        item = self.currentItem()
        self.newPageCore(item, name)

    def newPageCore(self, item, newPageName):
        pagePath = os.path.join(self.notebookPath, self.itemToPagePath(item))
        if not newPageName:
            dialog = ItemDialog(self)
            dialog.setPath(pagePath)
            if dialog.exec_():
                newPageName = dialog.editor.text()
        if newPageName:
            if hasattr(item, 'text'):
                pagePath = os.path.join(self.notebookPath, 
                                        pagePath + '/')
            if not QDir(pagePath).exists():
                QDir(self.notebookPath).mkdir(pagePath)
            fileName = pagePath + newPageName + self.settings.fileExt
            fh = QFile(fileName)
            fh.open(QIODevice.WriteOnly)
            savestream = QTextStream(fh)
            savestream << '# ' + newPageName + '\n'
            savestream << 'Created ' + str(datetime.date.today()) + '\n\n'
            fh.close()
            QTreeWidgetItem(item, [newPageName])
            newItem = self.pagePathToItem(pagePath + newPageName)
            self.sortItems(0, Qt.AscendingOrder)
            self.setCurrentItem(newItem)
            if hasattr(item, 'text'):
                self.expandItem(item)

            # TODO improvement needed, can be reused somehow
            fileobj = open(fileName, 'r')
            content = fileobj.read()
            fileobj.close()
            self.ix = open_dir(self.indexdir)
            writer = self.ix.writer()
            writer.add_document(path=pagePath+newPageName, content=content)
            writer.commit()

    def dropEvent(self, event):
        sourceItem = self.currentItem()
        sourcePath = self.itemToPagePath(sourceItem)
        targetItem = self.itemAt(event.pos())
        targetPath = self.itemToPagePath(targetItem)
        if targetPath != '':
            targetPath = targetPath + '/'
        oldName = self.noteFullPath(sourceItem.text(0))
        newName = targetPath + sourceItem.text(0) + self.settings.fileExt
        oldDir = sourcePath
        newDir = targetPath + sourceItem.text(0)
        print(newName)
        # if QDir(newName).exists():
        if QFile.exists(newName):
            QMessageBox.warning(self, 'Error',
                                'Page already exists: %s' % newName)
            return

        # if not QDir(newName).exists():
        QDir(self.notebookPath).mkpath(targetPath)
        QDir(self.notebookPath).rename(oldName, newName)
        if sourceItem.childCount() != 0:
            QDir(self.notebookPath).rename(oldDir, newDir)
        if sourceItem.parent() is not None:
            parentItem = sourceItem.parent()
            parentPath = self.itemToPagePath(parentItem)
            if parentItem.childCount() == 1:
                QDir(self.notebookPath).rmdir(parentPath)
        QTreeWidget.dropEvent(self, event)
        self.sortItems(0, Qt.AscendingOrder)
        if hasattr(targetItem, 'text'):
            self.expandItem(targetItem)

    def renamePageWrapper(self):
        item = self.currentItem()
        self.renamePage(item)

    def renamePage(self, item):
        parent = item.parent()
        parentPath = self.itemToPagePath(parent)
        dialog = ItemDialog(self)
        dialog.setPath(parentPath)
        dialog.setText(item.text(0))
        if dialog.exec_():
            newPageName = dialog.editor.text()
            # pagePath = self.getPath(item)
            # if hasattr(item, 'text'):       # if item is not QTreeWidget
            if parentPath != '':
                parentPath = parentPath + '/'
            oldName = self.noteFullPath(item.text(0))
            newName = parentPath + newPageName + self.settings.fileExt
            QDir(self.notebookPath).rename(oldName, newName)
            if item.childCount() != 0:
                oldDir = parentPath + item.text(0)
                newDir = parentPath + newPageName
                QDir(self.notebookPath).rename(oldDir, newDir)
            item.setText(0, newPageName)
            self.sortItems(0, Qt.AscendingOrder)

    def exists(self, noteFullName):
        if self.noteFullPath(noteFullName) != "" :
            return True
        else:
            return False

    def delPageWrapper(self):
        item = self.currentItem()
        self.delPage(item)

    def delPage(self, item):

        index = item.childCount()
        while index > 0:
            index = index - 1
            self.dirname = item.child(index).text(0)
            self.delPage(item.child(index))

        pagePath = self.itemToPagePath(item)
        self.ix = open_dir(self.indexdir)
        query = QueryParser('path', self.ix.schema).parse(pagePath)
        writer = self.ix.writer()
        n = writer.delete_by_query(query)
        # n = writer.delete_by_term('path', pagePath)
        writer.commit()
        b = QDir(self.notebookPath).remove(self.noteFullPath(pagePath))
        parent = item.parent()
        parentPath = self.itemToPagePath(parent)
        if parent is not None:
            index = parent.indexOfChild(item)
            parent.takeChild(index)
            if parent.childCount() == 0:  # if no child, dir not needed
                QDir(self.notebookPath).rmdir(parentPath)
        else:
            index = self.indexOfTopLevelItem(item)
            self.takeTopLevelItem(index)
        QDir(self.notebookPath).rmdir(pagePath)

    def recurseCollapse(self, item):
        for i in range(item.childCount()):
            a_item = item.child(i)
            self.recurseCollapse(a_item)
            self.collapseItem(item)

    def recurseExpand(self, item):
        self.expandItem(item)
        for i in range(item.childCount()):
            a_item = item.child(i)
            self.recurseExpand(a_item)



"""
Search module

Search results are displayed in a QWebView widget.
"""

from PyQt4.QtCore import QUrl
from PyQt4.QtGui import QCursor, QToolTip
from PyQt4.QtWebKit import QWebView, QWebPage


class MikiSearch(QWebView):

    def __init__(self, parent=None):
        super(MikiSearch, self).__init__(parent)
        self.parent = parent

        self.settings().clearMemoryCaches()
        self.flag = False
        self.link = None
        self.setMouseTracking(True)
        self.page().linkHovered.connect(self.linkHovered)
        self.page().setLinkDelegationPolicy(QWebPage.DelegateAllLinks)
        self.page().linkClicked.connect(self.linkClicked)

    def linkClicked(self, qurl):
        """ Overload function.
            Click link to open the note.
        """
        path = qurl.toString()
        item = self.parent.notesTree.pagePathToItem(path)
        if item:
            self.parent.notesTree.setCurrentItem(item)

    def linkHovered(self, link, title, textContent):
        """ Overload function.
            Show tooltip when hovered.
        # ToFix: tooltip disappear too soon.
        """
        #self.setToolTip(link)
        if link:
            self.flag = True
            self.link = link
            QToolTip.showText(QCursor.pos(), self.link)
        else:
            self.flag = False

    def mouseMoveEvent(self, event):
        if self.flag:
            QToolTip.showText(QCursor.pos(), self.link)
        else:
            QToolTip.hideText()
        QWebView.mouseMoveEvent(self, event)




"""
Mikidown tray icon module.
"""

from PyQt4.QtCore import Qt
from PyQt4.QtGui import QMenu, QSystemTrayIcon


class MikiTray(QSystemTrayIcon):
    def __init__(self, icon, parent=None):
        super(MikiTray, self).__init__(parent)
        self.parent = parent
        self.setIcon(icon)
        self.setVisible(True)

        menu = QMenu()
        menu.addAction(parent.actionQuit)
        self.setContextMenu(menu)
        self.activated.connect(self.toggleShow)

    def toggleShow(self, reason):
        """ Left click tray icon to toggle the display of MainWindow. 
        """
        if reason != QSystemTrayIcon.Trigger:
            return
        s = self.parent.windowState() 
        if self.parent.isVisible():
            if s == Qt.WindowMinimized:
                self.parent.showNormal()
                self.parent.show()
            else:
                self.parent.showMaximized
                self.parent.hide()
        else:
            self.parent.show()



from distutils import log
from distutils.core import setup
from distutils.command.build import build
from distutils.command.install_scripts import install_scripts
import glob
import sys

from mikidown.config import __version__

class miki_build(build):
    def run(self):
        # Check the python version
        try:
            version_info = sys.version_info
            assert version_info > (3, 0)
        except:
            print('ERROR: mikidown needs python >= 3.0', file=sys.stderr)
            sys.exit(1)
        build.run(self)

class miki_install_scripts(install_scripts):
    def run(self):
        import shutil
        install_scripts.run(self)
        for file in self.get_outputs():
            log.info('renaming %s to %s', file, file[:-3])
            shutil.move(file, file[:-3])

setup(
    name='mikidown',
    version=__version__,
    license = 'MIT',
    description = 'A note taking application, featuring markdown syntax',
    author = 'rnons',
    author_email = 'remotenonsense@gmail.com',
    url = 'https://github.com/rnons/mikidown',
    scripts=['mikidown.py'],
    packages=['mikidown'],
    data_files=[('share/mikidown', ['README.mkd']), 
                ('share/mikidown', ['mikidown/notes.css']), 
                ('share/mikidown/css', glob.glob("mikidown/css/*")), 
                ('share/icons/hicolor/scalable/apps', ['mikidown/icons/mikidown.svg']), 
                ('share/applications', ['mikidown.desktop'])
                ],
    requires=['PyQt', 'markdown', 'whoosh'],
    install_requires=['Markdown >= 2.3.1', 'Whoosh >= 2.5.2'],
    cmdclass={
        'build': miki_build,
        'install_scripts': miki_install_scripts
    },

    classifiers=[
        "Topic :: Text Editors :: Documentation",
        "Development Status :: 3 - Alpha",
        "Environment :: X11 Applications",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3"
    ]
)



import os
from multiprocessing import Process
from PyQt4.QtCore import QDir, QFile, QFileInfo, QTextStream, QIODevice
from PyQt4.QtGui import QTextEdit, QFileDialog, QMessageBox
import markdown
from whoosh.index import open_dir

from mikidown.config import *
from mikidown.utils import parseTitle


class MikiEdit(QTextEdit):

    def __init__(self, parent=None):
        super(MikiEdit, self).__init__(parent)
        self.settings = parent.settings
        self.setFontPointSize(12)
        self.setTabStopWidth(4)
        self.setVisible(False)
        indexdir = os.path.join(self.settings.notebookPath, 
                                self.settings.indexdir)
        self.ix = open_dir(indexdir)

    def updateIndex(self, path, content):
            ''' Update whoosh index, which cost much computing resource '''
            writer = self.ix.writer()
            writer.update_document(
                path=path, title=parseTitle(content, path), content=content)
            writer.commit()

    def save(self, pageName, filePath):
        fh = QFile(filePath)
        try:
            if not fh.open(QIODevice.WriteOnly):
                raise IOError(fh.errorString())
        except IOError as e:
            QMessageBox.warning(self, 'Save Error',
                                'Failed to save %s: %s' % (pageName, e))
        finally:
            if fh is not None:
                savestream = QTextStream(fh)
                savestream << self.toPlainText()
                fh.close()
                self.document().setModified(False)
                # Fork a process to update index, which benefit responsiveness.
                p = Process(target=self.updateIndex, args=(
                    pageName, self.toPlainText(),))
                p.start()

    def toHtml(self):
        '''markdown.Markdown.convert v.s. markdown.markdown
            Previously `convert` was used, but it doens't work with fenced_code
        '''
        htmltext = self.toPlainText()
        return markdown.markdown(htmltext, self.settings.extensions)
        # md = markdown.Markdown(extensions)
        # return md.convert(htmltext)

    def saveAsHtml(self):
        """ Export current note as html file
        """
        fileName = QFileDialog.getSaveFileName(self, self.tr('Export to HTML'),
                                               '', '(*.html *.htm);;'+self.tr('All files(*)'))
        if fileName == '':
            return
        if not QFileInfo(fileName).suffix():
            fileName += '.html'
        html = QFile(fileName)
        html.open(QIODevice.WriteOnly)
        savestream = QTextStream(html)
        css = QFile(self.settings.notebookPath + '/notes.css')
        css.open(QIODevice.ReadOnly)
        # Use a html lib may be a better idea?
        savestream << "<html><head><meta charset='utf-8'></head>"
        # Css is inlined.
        savestream << "<style>"
        savestream << QTextStream(css).readAll()
        savestream << "</style>"
        # Note content
        savestream << self.toHtml()
        savestream << "</html>"
        html.close()
